==> xl-plugin-manager-24.1.10-sources.jar: com/xebialabs/plugin/manager/service/PluginService.scala <==
package com.xebialabs.plugin.manager.service

import com.xebialabs.deployit.engine.spi.exception.{DeployitException, HttpResponseCodeResult}
import com.xebialabs.deployit.exception.NotFoundException
import com.xebialabs.plugin.manager.PluginId.{Artifact, LocalFile}
import com.xebialabs.plugin.manager.compatibility.Jdk17Compatibility
import com.xebialabs.plugin.manager.config.PluginManagerProperties
import com.xebialabs.plugin.manager.event.{PluginInstalledEvent, PluginUpdatedEvent}
import com.xebialabs.plugin.manager.metadata.{ArtifactId, PluginMetadata}
import com.xebialabs.plugin.manager.model.DbPlugin
import com.xebialabs.plugin.manager.repository.PluginsRepository
import com.xebialabs.plugin.manager.repository.nexus.{NexusPluginRepository, NexusRepositoryConfig, NexusServerConfig}
import com.xebialabs.plugin.manager.rest.api.{PluginSource, PluginStatus}
import com.xebialabs.plugin.manager.rest.dto.PluginDto
import com.xebialabs.plugin.manager.service.PluginService.defaultTimeout
import com.xebialabs.plugin.manager.startup.PluginReloader
import com.xebialabs.plugin.manager.validator.{PluginValidator, TypeSystemValidator}
import com.xebialabs.plugin.manager.{Plugin, PluginId, PluginManager}
import grizzled.slf4j.Logging
import jakarta.annotation.PreDestroy
import org.springframework.context.ApplicationEventPublisher

import java.io.File
import scala.collection.mutable
import scala.concurrent.duration._
import scala.concurrent.{Await, ExecutionContext, Future}
import scala.jdk.CollectionConverters._
import scala.language.postfixOps
import scala.util.{Failure, Success, Try}

trait PluginService extends Logging {
  def typeSystemValidator: TypeSystemValidator

  def pluginManager: PluginManager

  def pluginReloader: PluginReloader

  def eventPublisher: ApplicationEventPublisher

  implicit lazy val ec: ExecutionContext = pluginManager.ec

  def repositories: mutable.Map[String, PluginsRepository]

  def addRepository(repository: PluginsRepository): Boolean

  def deleteRepository(name: String): Boolean

  def shouldCheckJdk17Compatibility: Boolean


  def update(): Unit = {
    if (repositories.nonEmpty) {
      Await.ready(Future.sequence(repositories.values.map(_.update())), defaultTimeout)
    }
  }

  def search(query: Option[String]): Map[ArtifactId, Option[PluginMetadata]] = {
    logger.debug(s"search($query)")
    Await.result(
      for {
        installed <- pluginManager.search(query)
        available <- Future.sequence(repositories.values.map(_.search(query))).map(_.flatten)
      } yield {
        available.foldLeft(installed.map(_ -> Option.empty[PluginMetadata]).toMap) {
          case (acc, (id, pm1)) =>
            acc.get(id) match {
              case None =>
                acc.updated(id, pm1)
              case Some(_) =>
                pm1.map(pm => acc.updated(id, Some(pm))).getOrElse(acc)
            }
        }
      },
      defaultTimeout
    )
  }

  def searchOfficial(query: String): Map[ArtifactId, Option[PluginMetadata]] = {
    Await.result(
      for {
        available <- Future.sequence(repositories.values.map(_.search(Some(query)))).map(_.flatten)
      } yield {
        available.toMap
      },
      defaultTimeout
    )
  }

  def search(query: String): Map[ArtifactId, Option[PluginMetadata]] = search(Option(query).filter(_.nonEmpty))

  def list(): Map[ArtifactId, Option[PluginMetadata]] = search(None)

  def listInstalled(): Seq[PluginId] =
    pluginManager.listInstalled()

  def installOrUpdate(plugin: Plugin): Unit = {
    logger.info(s"Installing plugin ${plugin.id.id}...")
    pluginManager.installOrUpdate(plugin)
    logger.info(s"Installation of plugin ${plugin.id.id} to the database is complete. " +
      s"System restart required for the plugin to be ready for usage.")
  }

  def install(pluginId: PluginId, bytes: Array[Byte]): InstallationResult = {
    logger.info(s"Installing plugin $pluginId...")
    if (shouldCheckJdk17Compatibility) {
      Try(Jdk17Compatibility.isCompatible(s"${pluginId.name}.${pluginId.extension}", bytes)) match {
        case Success(false) =>
          logger.warn(s"Installation of plugin ${pluginId.name} failed due to JDK17 incompatibility.")
          return InstallationFailure(pluginId.name, s"${pluginId.name} is not JDK17 compatible.")
        case Success(true) =>
        case Failure(ex) =>
          logger.warn(s"Installation of plugin ${pluginId.name} failed due to JDK17 incompatibility.")
          return InstallationFailure(pluginId.name, s"Unable to verify plugin ${pluginId.name} JDK17 compatibility. Error message ${ex.getMessage}")
      }
    }

    validateInstall(pluginId, bytes) {
      val plugin = Plugin(pluginId, None, bytes)
      val result = doWithTypeSystemValidation(plugin) {
        pluginManager.install(plugin)
      }
      publishPluginManagerEvent(result)
      result
    }
  }

  private def validateInstall(pluginId: PluginId, bytes: Array[Byte])(successCallback: => InstallationResult): InstallationResult = {
    val installedPlugins = listInstalled()
    val sameVersionAlreadyInstalled = installedPlugins.exists(_.equals(pluginId))
    val pluginName = pluginId.name
    val differentVersionAlreadyInstalled = installedPlugins.exists(_.name.equals(pluginName))

    if (sameVersionAlreadyInstalled) {
      val result = SameVersionInstalledAlready(pluginName)
      logger.info(result.message)
      result
    } else if (differentVersionAlreadyInstalled) {
      val result = DifferentVersionInstalledAlready(pluginName)
      logger.info(result.message)
      result
    } else {
      val valid = new PluginValidator().validate(bytes, pluginId)
      if (valid.success) successCallback else valid
    }
  }

  def update(existingPluginName: String, newVersion: PluginId.LocalFile, bytes: Array[Byte]): InstallationResult = {
    logger.info(s"Updating plugin $existingPluginName to version $newVersion...")
    val matchingPlugins = listInstalled().filter(plugin => plugin.pluginId.name.startsWith(existingPluginName) && plugin.pluginId.isInstanceOf[LocalFile])
    if (matchingPlugins.size > 1) {
      val result = MultipleInstalledPluginsFoundForUpdate(existingPluginName)
      logger.info(result.message)
      result
    } else if (matchingPlugins.isEmpty) {
      val result = ZeroInstalledPluginsFoundForUpdate(existingPluginName)
      logger.info(result.message)
      result
    } else {
      if (shouldCheckJdk17Compatibility) {
        Try(Jdk17Compatibility.isCompatible(existingPluginName, bytes)) match {
          case Success(false) =>
            logger.warn(s"Update of plugin $existingPluginName failed due to JDK17 incompatibility.")
            return UpdateFailed(existingPluginName, s"$existingPluginName is not JDK17 compatible.")
          case Success(true) =>
          case Failure(ex) =>
            logger.warn(s"Update of plugin $existingPluginName failed due to JDK17 incompatibility.")
            return UpdateFailed(existingPluginName, s"Unable to verify plugin $existingPluginName JDK17 compatibility. Error message ${ex.getMessage}")
        }
      }

      val alreadyInstalled = matchingPlugins.head.asInstanceOf[LocalFile]
      val newPlugin = Plugin(newVersion, None, bytes)
      val result = doWithTypeSystemValidation(newPlugin) {
        pluginManager.update(alreadyInstalled, newPlugin)
      }
      publishPluginManagerEvent(result)
      result
    }
  }

  def installOrUpdateFromRepository(id: PluginId.Artifact): Try[Unit] =
    repositories.get(id.repository).map { repo =>
      Try {
        val plugin = Await.result(repo.get(id), defaultTimeout)
        val installationResult = doWithTypeSystemValidation(plugin) {
          pluginManager.install(plugin)
        }
        publishPluginManagerEvent(installationResult)
        if (!installationResult.success) {
          throw new IllegalStateException(installationResult.message)
        }
      }.recoverWith {
        case err =>
          logger.warn(err.getMessage)
          Failure(err)
      }
    }.getOrElse {
      Failure(new NotFoundException(s"Unknown plugin repository '${id.repository}"))
    }

  private def doWithTypeSystemValidation(plugin: Plugin)(pluginOperation: => Unit): InstallationResult = {
    val validations = typeSystemValidator.validate(plugin)

    validations.messages.foreach(info => logger.info(info))
    validations.warnings.foreach(warning => logger.warn(warning))
    validations.errors.foreach(error => logger.error(error))

    if (validations.failed()) {
      InstallationFailure(plugin.id.name, validations.messages.mkString("\n"), validations.warnings, validations.errors)
    } else {
      // requires a restart - i.e. it's a normal InstallationResult
      val result = Try(pluginOperation) match {
        case Failure(exception) =>
          logger.warn(s"Unable install/update ${plugin.id.name}", exception)
          InstallationFailure(plugin.id.name, exception.getMessage, validations.warnings, validations.errors)
        case Success(_) =>
          InstallationSuccess(plugin.id.name, requiresRestart = true, validations.warnings)
      }
      if (validations.restartRequired) {
        result
      } else {
        // broadcast ReloadPlugins message which is going to invoke pluginReloader.reloadPlugin on all of the nodes
        pluginReloader.reloadPlugin(plugin)
        InstallationSuccess(plugin.id.name, requiresRestart = false, validations.warnings)
      }
    }
  }


  def installFromRepository(id: PluginId.Artifact): InstallationResult = {
    try {
      repositories.get(id.repository).map { repo =>
        val plugin = Await.result(repo.get(id), defaultTimeout)
        validateInstall(plugin.id, plugin.bytes) {
          val result = doWithTypeSystemValidation(plugin) {
            pluginManager.install(plugin)
          }
          publishPluginManagerEvent(result)
          result
        }
      }.getOrElse {
        val result = RepositoryNotFound(id.repository, id.name)
        logger.info(result.message)
        result
      }
    } catch {
      case err: Exception =>
        logger.warn(err.getMessage)
        val result = PluginNotFound(id.name)
        logger.info(result.message)
        result
    }
  }

  def updateFromRepository(id: PluginId.Artifact): InstallationResult = {
    def doUpdate(repo: PluginsRepository, newVersion: PluginId.Artifact, existing: PluginId): InstallationResult = {
      val newPlugin = Await.result(repo.get(newVersion), defaultTimeout)
      doWithTypeSystemValidation(newPlugin) {
        pluginManager.update(existing, newPlugin)
      }
    }

    repositories.get(id.repository).map { repo =>
      val installedPlugins = getPluginsToUpdate(id.name)
      val availablePlugins = getUpdateMatchingPlugins(id)

      val updateResult = UpdateStrategy.doUpdate(repo, installedPlugins, availablePlugins, id.name, doUpdate)
      logger.info(updateResult.message)
      publishPluginManagerEvent(updateResult)
      updateResult
    }.getOrElse {
      val result = RepositoryNotFound(id.repository, id.name)
      logger.info(result.message)
      result
    }
  }

  def getPluginsToUpdate(pluginName: String): Seq[PluginId] = {
    val pluginNameCriteria = if (pluginName.endsWith("*")) pluginName.dropRight(1) else pluginName
    listInstalled().filter(plugin => plugin.pluginId.name.startsWith(pluginNameCriteria) && plugin.pluginId.isInstanceOf[Artifact])
  }

  def getUpdateMatchingPlugins(id: PluginId.Artifact): Map[ArtifactId, Option[PluginMetadata]] = {
    if (id.name.endsWith("*")) {
      searchOfficial(id.name.dropRight(1))
    } else {
      searchOfficial(id.name)
        .filter(item => item._1.repository.get.equals(id.repository) && item._1.groupId.equals(id.groupId) && item._1.artifactId.equals(id.artifactId))
    }
  }

  def uninstall(repositoryId: PluginSource.Value,
                groupId: String,
                artifactId: String,
                version: Option[String]): Try[Unit] = {
    logger.info(s"Uninstalling plugin $repositoryId:$groupId:$artifactId-$version")

    pluginManager.getPluginBy(repositoryId, groupId, artifactId, version) match {
      case Some(plugin) =>
        if (pluginManager.uninstall(plugin)) {
          pluginManager.revertPreviousVersionIfNecessary(plugin)
          Success((): Unit)
        } else {
          Failure(PluginUninstallException(plugin, s"Only plugins in ${PluginStatus.READY_FOR_INSTALL} status can be uninstalled."))
        }
      case None =>
        Failure(PluginNotFoundException(s"Plugin $repositoryId:$groupId:$artifactId-$version not found in the system."))
    }
  }

  def getLogo(repositoryId: String, groupId: String, artifactId: String): Option[File] =
    for {
      repo <- repositories.get(repositoryId)
      logo <- repo.getLogo(groupId, artifactId)
    } yield logo

  def attachMetadata(pluginId: PluginId): PluginDto = {
    val repoId = pluginId.source.toString
    val metadata = for {
      repo <- repositories.get(repoId)
      pluginsMeta <- repo.getMetadata(pluginId.toArtifactId)
    } yield pluginsMeta
    PluginDto(pluginId, metadata)
  }

  def attachStatus: PluginDto => PluginDto = {
    val statuses: Map[String, PluginStatus.Value] = pluginManager.fetchPluginStatuses()
    pluginDto => pluginDto.copy(status = statuses.get(pluginDto.plugin.name))
  }

  def extend(data: Seq[PluginId]): Seq[PluginDto] = data.map(attachMetadata).map(attachStatus)

  @PreDestroy
  def shutdown(): Unit = {
    logger.info("Shutting down LocalPluginManager...")
    Await.ready(
      Future.sequence(repositories.values.collect {
        case nexus: NexusPluginRepository => nexus.shutdown()
      }),
      repositories.size * defaultTimeout
    )
  }

  private def publishPluginManagerEvent(result: InstallationResult): Unit = {
    result match {
      case InstallationSuccess(pluginName, requiresRestart, _) => eventPublisher.publishEvent(PluginInstalledEvent(this, pluginName, requiresRestart))
      case UpdateSuccess(pluginName, requiresRestart) => eventPublisher.publishEvent(PluginUpdatedEvent(this, pluginName, requiresRestart))
      case _ => ()
    }
  }

}

object PluginService {
  // TODO: get this from configuration
  val defaultTimeout: Duration = 180 seconds

  def serversFromConfig(pluginsConfig: PluginManagerProperties): Map[String, Try[NexusServerConfig]] =
    Option(pluginsConfig.getServers) match {
      case Some(servers) => servers.asScala.collect(server =>
        server.getName -> NexusServerConfig.fromConfig(server)).toMap
      case None => Map.empty
    }

  def repositoriesFromConfig(pluginsConfig: PluginManagerProperties, servers: Map[String, Try[NexusServerConfig]])
                            (implicit productConfig: ProductConfig): Map[String, Try[NexusRepositoryConfig]] =
    Option(pluginsConfig.getRepositories) match {
      case Some(repositories) => repositories.asScala.map(repo =>
        repo.getName -> NexusRepositoryConfig.fromConfig(repo.getName)(repo, servers)).toMap
      case None => Map.empty
    }

  def configuredRepositories(pluginsConfig: PluginManagerProperties)
                            (implicit productConfig: ProductConfig): List[Try[NexusPluginRepository]] =
    repositoriesFromConfig(pluginsConfig, serversFromConfig(pluginsConfig))
      .map { case (name, tryConfig) =>
        tryConfig.map(config => NexusPluginRepository.memCached(name, config))
      }.toList
}


@HttpResponseCodeResult(statusCode = 501)
case class PluginUninstallException(plugin: DbPlugin, message: String) extends DeployitException(s"Unable to uninstall plugin $plugin. $message")

@HttpResponseCodeResult(statusCode = 404)
case class PluginNotFoundException(message: String) extends DeployitException(message)

==> xl-plugin-manager-25.1.6-1005.800-sources.jar: com/xebialabs/plugin/manager/service/PluginService.scala <==
package com.xebialabs.plugin.manager.service

import com.xebialabs.deployit.engine.spi.exception.{DeployitException, HttpResponseCodeResult}
import com.xebialabs.deployit.exception.NotFoundException
import com.xebialabs.plugin.manager.PluginId.{Artifact, LocalFile}
import com.xebialabs.plugin.manager.compatibility.Jdk17Compatibility
import com.xebialabs.plugin.manager.config.PluginManagerProperties
import com.xebialabs.plugin.manager.event.{PluginInstalledEvent, PluginUpdatedEvent}
import com.xebialabs.plugin.manager.metadata.{ArtifactId, PluginMetadata}
import com.xebialabs.plugin.manager.model.DbPlugin
import com.xebialabs.plugin.manager.repository.PluginsRepository
import com.xebialabs.plugin.manager.repository.artifactory.{ArtifactoryPluginRepository, ArtifactoryRepositoryConfig, ArtifactoryServerConfig}
import com.xebialabs.plugin.manager.repository.config.{PluginRepositoryConfig, PluginServerConfig}
import com.xebialabs.plugin.manager.repository.nexus.{NexusPluginRepository, NexusRepositoryConfig, NexusServerConfig}
import com.xebialabs.plugin.manager.rest.api.{PluginSource, PluginStatus}
import com.xebialabs.plugin.manager.rest.dto.PluginDto
import com.xebialabs.plugin.manager.service.PluginService.defaultTimeout
import com.xebialabs.plugin.manager.startup.PluginReloader
import com.xebialabs.plugin.manager.validator.{PluginValidator, TypeSystemValidator}
import com.xebialabs.plugin.manager.{Plugin, PluginId, PluginManager}
import grizzled.slf4j.Logging
import jakarta.annotation.PreDestroy
import org.springframework.context.ApplicationEventPublisher

import java.io.File
import scala.collection.mutable
import scala.concurrent.duration._
import scala.concurrent.{Await, ExecutionContext, Future}
import scala.jdk.CollectionConverters._
import scala.language.postfixOps
import scala.util.{Failure, Success, Try}

trait PluginService extends Logging {
  def typeSystemValidator: TypeSystemValidator

  def pluginManager: PluginManager

  def pluginReloader: PluginReloader

  def eventPublisher: ApplicationEventPublisher

  implicit lazy val ec: ExecutionContext = pluginManager.ec

  def repositories: mutable.Map[String, PluginsRepository]

  def addRepository(repository: PluginsRepository): Boolean

  def deleteRepository(name: String): Boolean

  def shouldCheckJdk17Compatibility: Boolean


  def update(): Unit = {
    if (repositories.nonEmpty) {
      Await.ready(Future.sequence(repositories.values.map(_.update())), defaultTimeout)
    }
  }

  def search(query: Option[String]): Map[ArtifactId, Option[PluginMetadata]] = {
    logger.debug(s"search($query)")
    Await.result(
      for {
        installed <- pluginManager.search(query)
        available <- Future.sequence(repositories.values.map(_.search(query))).map(_.flatten)
      } yield {
        available.foldLeft(installed.map(_ -> Option.empty[PluginMetadata]).toMap) {
          case (acc, (id, pm1)) =>
            acc.get(id) match {
              case None =>
                acc.updated(id, pm1)
              case Some(_) =>
                pm1.map(pm => acc.updated(id, Some(pm))).getOrElse(acc)
            }
        }
      },
      defaultTimeout
    )
  }

  def searchOfficial(query: String): Map[ArtifactId, Option[PluginMetadata]] = {
    Await.result(
      for {
        available <- Future.sequence(repositories.values.map(_.search(Some(query)))).map(_.flatten)
      } yield {
        available.toMap
      },
      defaultTimeout
    )
  }

  def search(query: String): Map[ArtifactId, Option[PluginMetadata]] = search(Option(query).filter(_.nonEmpty))

  def list(): Map[ArtifactId, Option[PluginMetadata]] = search(None)

  def listInstalled(): Seq[PluginId] =
    pluginManager.listInstalled()

  def installOrUpdate(plugin: Plugin): Unit = {
    logger.info(s"Installing plugin ${plugin.id.id}...")
    pluginManager.installOrUpdate(plugin)
    logger.info(s"Installation of plugin ${plugin.id.id} to the database is complete. " +
      s"System restart required for the plugin to be ready for usage.")
  }

  def install(pluginId: PluginId, bytes: Array[Byte]): InstallationResult = {
    logger.info(s"Installing plugin $pluginId...")
    if (shouldCheckJdk17Compatibility) {
      Try(Jdk17Compatibility.isCompatible(s"${pluginId.name}.${pluginId.extension}", bytes)) match {
        case Success(false) =>
          logger.warn(s"Installation of plugin ${pluginId.name} failed due to JDK17 incompatibility.")
          return InstallationFailure(pluginId.name, s"${pluginId.name} is not JDK17 compatible.")
        case Success(true) =>
        case Failure(ex) =>
          logger.warn(s"Installation of plugin ${pluginId.name} failed due to JDK17 incompatibility.")
          return InstallationFailure(pluginId.name, s"Unable to verify plugin ${pluginId.name} JDK17 compatibility. Error message ${ex.getMessage}")
      }
    }

    validateInstall(pluginId, bytes) {
      val plugin = Plugin(pluginId, None, bytes)
      val result = doWithTypeSystemValidation(plugin) {
        pluginManager.install(plugin)
      }
      publishPluginManagerEvent(result)
      result
    }
  }

  private def validateInstall(pluginId: PluginId, bytes: Array[Byte])(successCallback: => InstallationResult): InstallationResult = {
    val installedPlugins = listInstalled()
    val sameVersionAlreadyInstalled = installedPlugins.exists(_.equals(pluginId))
    val pluginName = pluginId.name
    val differentVersionAlreadyInstalled = installedPlugins.exists(_.name.equals(pluginName))

    if (sameVersionAlreadyInstalled) {
      val result = SameVersionInstalledAlready(pluginName)
      logger.info(result.message)
      result
    } else if (differentVersionAlreadyInstalled) {
      val result = DifferentVersionInstalledAlready(pluginName)
      logger.info(result.message)
      result
    } else {
      val valid = new PluginValidator().validate(bytes, pluginId)
      if (valid.success) successCallback else valid
    }
  }

  def update(existingPluginName: String, newVersion: PluginId.LocalFile, bytes: Array[Byte]): InstallationResult = {
    logger.info(s"Updating plugin $existingPluginName to version $newVersion...")
    val matchingPlugins = listInstalled().filter(plugin => plugin.pluginId.name.startsWith(existingPluginName) && plugin.pluginId.isInstanceOf[LocalFile])
    if (matchingPlugins.size > 1) {
      val result = MultipleInstalledPluginsFoundForUpdate(existingPluginName)
      logger.info(result.message)
      result
    } else if (matchingPlugins.isEmpty) {
      val result = ZeroInstalledPluginsFoundForUpdate(existingPluginName)
      logger.info(result.message)
      result
    } else {
      if (shouldCheckJdk17Compatibility) {
        Try(Jdk17Compatibility.isCompatible(existingPluginName, bytes)) match {
          case Success(false) =>
            logger.warn(s"Update of plugin $existingPluginName failed due to JDK17 incompatibility.")
            return UpdateFailed(existingPluginName, s"$existingPluginName is not JDK17 compatible.")
          case Success(true) =>
          case Failure(ex) =>
            logger.warn(s"Update of plugin $existingPluginName failed due to JDK17 incompatibility.")
            return UpdateFailed(existingPluginName, s"Unable to verify plugin $existingPluginName JDK17 compatibility. Error message ${ex.getMessage}")
        }
      }

      val alreadyInstalled = matchingPlugins.head.asInstanceOf[LocalFile]
      val newPlugin = Plugin(newVersion, None, bytes)
      val result = doWithTypeSystemValidation(newPlugin) {
        pluginManager.update(alreadyInstalled, newPlugin)
      }
      publishPluginManagerEvent(result)
      result
    }
  }

  def installOrUpdateFromRepository(id: PluginId.Artifact): Try[Unit] =
    repositories.get(id.repository).map { repo =>
      Try {
        val plugin = Await.result(repo.get(id), defaultTimeout)
        val installationResult = doWithTypeSystemValidation(plugin) {
          pluginManager.install(plugin)
        }
        publishPluginManagerEvent(installationResult)
        if (!installationResult.success) {
          throw new IllegalStateException(installationResult.message)
        }
      }.recoverWith {
        case err =>
          logger.warn(err.getMessage)
          Failure(err)
      }
    }.getOrElse {
      Failure(new NotFoundException(s"Unknown plugin repository '${id.repository}"))
    }

  private def doWithTypeSystemValidation(plugin: Plugin)(pluginOperation: => Unit): InstallationResult = {
    val validations = typeSystemValidator.validate(plugin)

    validations.messages.foreach(info => logger.info(info))
    validations.warnings.foreach(warning => logger.warn(warning))
    validations.errors.foreach(error => logger.error(error))

    if (validations.failed()) {
      InstallationFailure(plugin.id.name, validations.messages.mkString("\n"), validations.warnings, validations.errors)
    } else {
      // requires a restart - i.e. it's a normal InstallationResult
      val result = Try(pluginOperation) match {
        case Failure(exception) =>
          logger.warn(s"Unable install/update ${plugin.id.name}", exception)
          InstallationFailure(plugin.id.name, exception.getMessage, validations.warnings, validations.errors)
        case Success(_) =>
          InstallationSuccess(plugin.id.name, requiresRestart = true, validations.warnings)
      }
      if (validations.restartRequired) {
        result
      } else {
        // broadcast ReloadPlugins message which is going to invoke pluginReloader.reloadPlugin on all of the nodes
        pluginReloader.reloadPlugin(plugin)
        InstallationSuccess(plugin.id.name, requiresRestart = false, validations.warnings)
      }
    }
  }


  def installFromRepository(id: PluginId.Artifact): InstallationResult = {
    try {
      repositories.get(id.repository).map { repo =>
        val plugin = Await.result(repo.get(id), defaultTimeout)
        validateInstall(plugin.id, plugin.bytes) {
          val result = doWithTypeSystemValidation(plugin) {
            pluginManager.install(plugin)
          }
          publishPluginManagerEvent(result)
          result
        }
      }.getOrElse {
        val result = RepositoryNotFound(id.repository, id.name)
        logger.info(result.message)
        result
      }
    } catch {
      case err: Exception =>
        logger.warn(err.getMessage)
        val result = PluginNotFound(id.name)
        logger.info(result.message)
        result
    }
  }

  def updateFromRepository(id: PluginId.Artifact): InstallationResult = {
    def doUpdate(repo: PluginsRepository, newVersion: PluginId.Artifact, existing: PluginId): InstallationResult = {
      val newPlugin = Await.result(repo.get(newVersion), defaultTimeout)
      doWithTypeSystemValidation(newPlugin) {
        pluginManager.update(existing, newPlugin)
      }
    }

    repositories.get(id.repository).map { repo =>
      val installedPlugins = getPluginsToUpdate(id.name)
      val availablePlugins = getUpdateMatchingPlugins(id)

      val updateResult = UpdateStrategy.doUpdate(repo, installedPlugins, availablePlugins, id.name, doUpdate)
      logger.info(updateResult.message)
      publishPluginManagerEvent(updateResult)
      updateResult
    }.getOrElse {
      val result = RepositoryNotFound(id.repository, id.name)
      logger.info(result.message)
      result
    }
  }

  def getPluginsToUpdate(pluginName: String): Seq[PluginId] = {
    val pluginNameCriteria = if (pluginName.endsWith("*")) pluginName.dropRight(1) else pluginName
    listInstalled().filter(plugin => plugin.pluginId.name.startsWith(pluginNameCriteria) && plugin.pluginId.isInstanceOf[Artifact])
  }

  def getUpdateMatchingPlugins(id: PluginId.Artifact): Map[ArtifactId, Option[PluginMetadata]] = {
    if (id.name.endsWith("*")) {
      searchOfficial(id.name.dropRight(1))
    } else {
      searchOfficial(id.name)
        .filter(item => item._1.repository.get.equals(id.repository) && item._1.groupId.equals(id.groupId) && item._1.artifactId.equals(id.artifactId))
    }
  }

  def uninstall(repositoryId: PluginSource.Value,
                groupId: String,
                artifactId: String,
                version: Option[String]): Try[Unit] = {
    logger.info(s"Uninstalling plugin $repositoryId:$groupId:$artifactId-$version")

    pluginManager.getPluginBy(repositoryId, groupId, artifactId, version) match {
      case Some(plugin) =>
        if (pluginManager.uninstall(plugin)) {
          pluginManager.revertPreviousVersionIfNecessary(plugin)
          Success((): Unit)
        } else {
          Failure(PluginUninstallException(plugin, s"Only plugins in ${PluginStatus.READY_FOR_INSTALL} status can be uninstalled."))
        }
      case None =>
        Failure(PluginNotFoundException(s"Plugin $repositoryId:$groupId:$artifactId-$version not found in the system."))
    }
  }

  def getLogo(repositoryId: String, groupId: String, artifactId: String): Option[File] =
    for {
      repo <- repositories.get(repositoryId)
      logo <- repo.getLogo(groupId, artifactId)
    } yield logo

  def attachMetadata(pluginId: PluginId): PluginDto = {
    val repoId = pluginId.source.toString
    val metadata = for {
      repo <- repositories.get(repoId)
      pluginsMeta <- repo.getMetadata(pluginId.toArtifactId)
    } yield pluginsMeta
    PluginDto(pluginId, metadata)
  }

  def attachStatus: PluginDto => PluginDto = {
    val statuses: Map[String, PluginStatus.Value] = pluginManager.fetchPluginStatuses()
    pluginDto => pluginDto.copy(status = statuses.get(pluginDto.plugin.name))
  }

  def extend(data: Seq[PluginId]): Seq[PluginDto] = data.map(attachMetadata).map(attachStatus)

  @PreDestroy
  def shutdown(): Unit = {
    logger.info("Shutting down LocalPluginManager...")
    Await.ready(
      Future.sequence(repositories.values.collect {
        case nexus: NexusPluginRepository => nexus.shutdown()
        case artifactory: ArtifactoryPluginRepository => artifactory.shutdown()
      }),
      repositories.size * defaultTimeout
    )
  }

  private def publishPluginManagerEvent(result: InstallationResult): Unit = {
    result match {
      case InstallationSuccess(pluginName, requiresRestart, _) => eventPublisher.publishEvent(PluginInstalledEvent(this, pluginName, requiresRestart))
      case UpdateSuccess(pluginName, requiresRestart) => eventPublisher.publishEvent(PluginUpdatedEvent(this, pluginName, requiresRestart))
      case _ => ()
    }
  }

}

object PluginService {
  // TODO: get this from configuration
  val defaultTimeout: Duration = 180 seconds

  def serversFromConfig(pluginsConfig: PluginManagerProperties): Map[String, Try[PluginServerConfig]] =
    Option(pluginsConfig.getServers) match {
      case Some(servers) => servers.asScala.collect(server =>
        server.getServerType match {
          case "nexus" => server.getName -> NexusServerConfig.fromConfig(server)
          case "artifactory" => server.getName -> ArtifactoryServerConfig.fromConfig(server)
        }
      ).toMap
      case None => Map.empty
    }

  def repositoriesFromConfig(pluginsConfig: PluginManagerProperties, servers: Map[String, Try[PluginServerConfig]])
                            (implicit productConfig: ProductConfig): Map[String, Try[PluginRepositoryConfig]] = {
    Option(pluginsConfig.getRepositories) match {
      case Some(repositories) => repositories.asScala.collect {
        case repo =>
          val serverConfigs = pluginsConfig.getServers.asScala.map(server => server.getName -> server).toMap
          val serverType = {
            Option(repo.getServer).map(_.getServerType)
              .orElse(Option(repo.getServerRef).flatMap(serverConfigs.get).map(_.getServerType))
              .getOrElse("")
          }
          serverType match {
            case "nexus" => repo.getName -> NexusRepositoryConfig.fromConfig(repo.getName)(repo, servers)
            case "artifactory" => repo.getName -> ArtifactoryRepositoryConfig.fromConfig(repo.getName)(repo, servers)
            case "" => repo.getName -> Failure(new RuntimeException(s"Server information not found for repository ${repo.getName}"))
            case otherType => repo.getName -> Failure(new RuntimeException(s"Unsupported server type for repository ${repo.getName}: $otherType"))

          }
      }.toMap
      case None => Map.empty
    }
  }

  def configuredRepositories(pluginsConfig: PluginManagerProperties)
                            (implicit productConfig: ProductConfig): List[Try[PluginsRepository]] = {
    repositoriesFromConfig(pluginsConfig, serversFromConfig(pluginsConfig))
      .map { case (name, tryConfig) =>
        tryConfig.map {
          case config@(_: NexusRepositoryConfig) => NexusPluginRepository.memCached(name, config)
          case config@(_: ArtifactoryRepositoryConfig) => ArtifactoryPluginRepository.memCached(name, config)
        }
      }.toList
  }
}


@HttpResponseCodeResult(statusCode = 501)
case class PluginUninstallException(plugin: DbPlugin, message: String) extends DeployitException(s"Unable to uninstall plugin $plugin. $message")

@HttpResponseCodeResult(statusCode = 404)
case class PluginNotFoundException(message: String) extends DeployitException(message)
